import json
import os
import subprocess
import tempfile
from importlib import resources
from pathlib import Path
from typing import Annotated, Any

import httpx
import typer
from inflection import parameterize
from loguru import logger
from markdown import markdown
from openai import OpenAI
from weasyprint import HTML

from smart_letters.config import attach_settings
from smart_letters.exceptions import handle_abort
from smart_letters.format import terminal_message, simple_message


def generate_letter(
    api_key: str,
    posting_text: str,
    resume_text: str,
    example_text: str | None = None,
    reprompts: list[tuple[str, str]] | None = None,
) -> str:
    client = OpenAI(api_key=api_key)

    messages = []

    messages.append(
        dict(
            role="developer",
            content="""
                You are a professional resume writer and career coach. Your task is to
                generate a cover letter for a job application based on the provided job
                posting and resume. You will be given a job posting and a resume, and
                you need to create a tailored cover letter that highlights the
                candidate's qualifications and suitability for the position.
            """,
        ),
    )

    messages.append(
        dict(
            role="developer",
            content=f"""
                This is the job posting. It is html scraped from a job board. It
                contains a lot of html markup, but you should ignore that and focus on
                the text content.

                {posting_text}
            """,
        ),
    )

    messages.append(
        dict(
            role="developer",
            content=f"""
                Use this resume as you write the letter. It highlights the skills and
                experience of the candidate.

                {resume_text}
            """,
        ),
    )

    if example_text:
        messages.append(
            dict(
                role="developer",
                content=f"""
                    You should use this cover letter as reference for what I would like
                    the resulting cover letter to look like. It is a markdown document.

                    {example_text}
                """,
            ),
        )

    messages.append(
        dict(
            role="developer",
            content="""
                The cover letter should not sound like it was written by a machine. It
                should be direct, personal, and energetic. It should be professional,
                but informal in an almost daring way.

                Avoid sentences of the form, "With X, I will Y". Do not use any of the
                following words:

                - challenge
                - opportunity
                - leverage
                - passionate

                The letter should highlight the candidate's leadership, technical
                excellence, and ambition.

                Do not include a salutation or closing. Only include the body of the
                letter. Don't include any other text. Just print the body of the letter
                itself.

                Again, do not provide any friendly text before the letter. Do not
                explain what you are doing. Do not say "Here is your letter". Just print
                the damn letter.
            """,
        )
    )

    messages.append(dict(role="user", content="Please write me a nice letter!"))

    if reprompts is not None:
        for old_letter, user_feedback in reprompts:
            messages.append(dict(role="assistant", content=old_letter))
            messages.append(dict(role="user", content=user_feedback))

    kwargs = dict(
        model="gpt-4o",
        n=1,
        messages=messages,
    )
    logger.debug(
        f"Getting openai output using params: \n{json.dumps(kwargs, indent=2)}"
    )

    cmp = client.chat.completions.create(**kwargs)  # type: ignore
    text = cmp.choices[0].message.content
    assert text is not None

    return text


def assemble_letter(
    text: str,
    candidate_name: str,
    company: str | None,
    heading_text: str | None,
    sig_path: Path | None,
) -> str:
    parts = []

    if heading_text:
        parts.append(heading_text)

    if company:
        parts.append(f"To the Hiring Team at {company}:")
    else:
        parts.append("Dear Hiring Manager,")

    parts.append("")
    parts.append(text)
    parts.append("")
    parts.append("Best regards,")
    parts.append("")

    if sig_path:
        parts.append(f"![signature](file://{sig_path.absolute()})")
        parts.append("")

    parts.append(candidate_name)

    return "\n".join(parts)


def edit_letter(text: str) -> str:
    logger.debug("Editing generated letter")
    with tempfile.NamedTemporaryFile() as tmp_file:
        tmp_path = Path(tmp_file.name)
        tmp_path.write_text(text)
        subprocess.run([os.environ["EDITOR"], str(tmp_path)])
        return tmp_path.read_text()


def asset_path(file_name: str):
    return resources.files(f"{__package__}.assets") / file_name


def write_letter(
    text: str,
    filename_prefix: str,
    company: str | None,
    position: str | None,
    dump_html: bool,
) -> Path:
    logger.debug("Saving letter to file")
    name = filename_prefix
    if company:
        name += f"--{parameterize(company)}"
    if position:
        name += f"--{parameterize(position)}"
    pdf_path = Path(f"{name}.pdf")

    html_content = markdown(text)

    if dump_html:
        html_path = Path(f"{name}.html")
        logger.debug(f"Dumping html to {html_path}")
        html_path.write_text(html_content)

    css_paths = [asset_path("styles.css")]
    html = HTML(string=html_content)
    html.write_pdf(pdf_path, stylesheets=css_paths)
    return pdf_path


cli = typer.Typer()


@cli.callback(invoke_without_command=True)
@handle_abort
@attach_settings
def generate(
    ctx: typer.Context,
    posting_url: Annotated[str, typer.Argument(help="The URL of the job posting.")],
    company: Annotated[
        str | None, typer.Option(help="The name of the company.")
    ] = None,
    position: Annotated[str | None, typer.Option(help="The title for the job.")] = None,
    example_letter: Annotated[
        Path | None, typer.Option(help="An example letter to use as a reference.")
    ] = None,
    dump_html: Annotated[
        bool, typer.Option(help="[FOR DEBUGGING] Dump generated HTML to a file.")
    ] = False,
    fake: Annotated[
        bool,
        typer.Option(
            help="[FOR DEBUGGING] Use a fake letter body instead of fetching one from OpenAI."
        ),
    ] = False,
):
    settings = ctx.obj.settings
    resume_path: Path = settings.resume_path
    candidate_name = settings.candidate_name
    filename_prefix = settings.filename_prefix
    heading_path = settings.heading_path
    sig_path = settings.sig_path

    logger.debug(f"Loading resume from {resume_path}")
    resume_text = resume_path.read_text()

    heading_text: str | None = None
    if heading_path:
        logger.debug(f"Loading heading text from {heading_path}")
        heading_text = heading_path.read_text()

    example_text: str | None = None
    if example_letter:
        logger.debug(f"Loading example text from {example_letter}")
        heading_text = example_letter.read_text()

    logger.debug(f"Pulling posting from {posting_url}")
    response = httpx.get(posting_url)
    response.raise_for_status()
    posting_text = response.text

    kwargs: dict[str, Any] = dict()
    if example_letter:
        logger.debug(f"Loading example letter from {example_letter}")
        kwargs["example_text"] = example_letter.read_text()

    logger.debug("Generating letter")

    api_key = ctx.obj.settings.openai_api_key

    generate_args: list[Any] = [api_key, candidate_name, posting_text, resume_text]
    generate_kwargs: dict[str, Any] = dict(
        example_text=example_text,
        sig_path=sig_path,
    )

    assemble_args = [
        candidate_name,
        company,
        heading_text,
        sig_path,
    ]

    if fake:
        text = asset_path("fake.txt").read_text()
    else:
        text = generate_letter(*generate_args, **generate_kwargs)

    full_text = assemble_letter(text, *assemble_args)

    accepted = False
    reprompts = []

    while not accepted:
        terminal_message(full_text, subject="Generated Letter", markdown=True)

        accepted = typer.confirm("Are you satisfied with the letter?", default=True)

        if not accepted:
            reprompts.append(
                (
                    text,
                    typer.prompt("What can I do to fix it?", default="Just try again"),
                )
            )
            generate_kwargs["reprompts"] = reprompts

            logger.debug("Regenerating letter based on feedback")
            text = generate_letter(*generate_args, **kwargs)
            full_text = assemble_letter(text, *assemble_args)

    edit = typer.confirm("Would you like to edit the letter?", default=False)
    if edit:
        edit_letter(full_text)

    pdf_path = write_letter(full_text, filename_prefix, company, position, dump_html)

    simple_message(f"Cover letter saved to {pdf_path}")



import tomllib
from importlib import metadata


def get_version_from_metadata() -> str:
    return metadata.version(__package__ or __name__)


def get_version_from_pyproject() -> str:
    with open("pyproject.toml", "rb") as file:
        return tomllib.load(file)["project"]["version"]


def get_version() -> str:
    try:
        return get_version_from_metadata()
    except metadata.PackageNotFoundError:
        try:
            return get_version_from_pyproject()
        except (FileNotFoundError, KeyError):
            return "unknown"


def show_version() -> None:
    print(get_version())



from contextlib import contextmanager
import json
from functools import wraps
from pathlib import Path
from typing import Annotated, Any

import snick
import typer
from loguru import logger
from pydantic import AfterValidator, BaseModel, ValidationError

from smart_letters.exceptions import Abort, handle_abort
from smart_letters.cache import CACHE_DIR, init_cache
from smart_letters.format import terminal_message


settings_path: Path = CACHE_DIR / "settings.json"


def file_exists(value: Path | None) -> Path | None:
    if value is None:
        return value

    value = value.expanduser()
    if not value.exists():
        raise ValueError(f"File not found at {value}")
    return value


class Settings(BaseModel):
    openai_api_key: str
    resume_path: Annotated[Path, AfterValidator(file_exists)]
    candidate_name: str
    filename_prefix: str = "cover-letter"
    heading_path: Annotated[Path | None, AfterValidator(file_exists)] = None
    sig_path: Annotated[Path | None, AfterValidator(file_exists)] = None


@contextmanager
def handle_config_error():
    try:
        yield
    except ValidationError as err:
        raise Abort(
            snick.conjoin(
                "A configuration error was detected.",
                "",
                "Details:",
                "",
                f"[red]{err}[/red]",
            ),
            subject="Configuration Error",
            log_message="Configuration error",
        )


def init_settings(**settings_values) -> Settings:
    with handle_config_error():
        logger.debug("Validating settings")
        return Settings(**settings_values)


def update_settings(settings: Settings, **settings_values) -> Settings:
    with handle_config_error():
        logger.debug("Validating settings")
        return settings.model_copy(update=settings_values)


def unset_settings(settings: Settings, *unset_keys) -> Settings:
    with handle_config_error():
        logger.debug("Unsetting settings")
        return Settings(
            **{
                k: v
                for (k, v) in settings.model_dump(exclude_unset=True).items()
                if k not in unset_keys
            }
        )


def attach_settings(func):
    @wraps(func)
    def wrapper(ctx: typer.Context, *args, **kwargs):
        try:
            logger.debug(f"Loading settings from {settings_path}")
            settings_values = json.loads(settings_path.read_text())
        except FileNotFoundError:
            raise Abort(
                f"""
                No settings file found at {settings_path}!

                Run the set-config sub-command first to establish your settings.
                """,
                subject="Settings file missing!",
                log_message="Settings file missing!",
            )
        logger.debug("Binding settings to CLI context")
        ctx.obj.settings = init_settings(**settings_values)
        return func(ctx, *args, **kwargs)

    return wrapper


def dump_settings(settings: Settings):
    logger.debug(f"Saving settings to {settings_path}")
    settings_values = settings.model_dump_json(indent=2)
    settings_path.write_text(settings_values)


def clear_settings():
    logger.debug(f"Removing saved settings at {settings_path}")
    settings_path.unlink(missing_ok=True)


cli = typer.Typer()


@cli.command()
@handle_abort
@init_cache
def bind(
    openai_api_key: Annotated[
        str, typer.Option(help="The API key needed to access OpenAI")
    ],
    resume_path: Annotated[Path, typer.Option(help="The path to your resume")],
    candidate_name: Annotated[
        str, typer.Option(help="The name of the candidate to use in the closing")
    ],
    filename_prefix: Annotated[
        str, typer.Option(help="The filename prefix to use for your cover letter")
    ] = "cover-letter",
    sig_path: Annotated[
        Path | None, typer.Option(help="The path to your signature")
    ] = None,
    heading_path: Annotated[
        Path | None, typer.Option(help="The path to your markdown heading")
    ] = None,
):
    """
    Bind the configuration to the app.
    """
    logger.debug(f"Initializing settings with {locals()}")
    settings = init_settings(
        openai_api_key=openai_api_key,
        resume_path=resume_path,
        candidate_name=candidate_name,
        filename_prefix=filename_prefix,
        sig_path=sig_path,
        heading_path=heading_path,
    )
    dump_settings(settings)


@cli.command()
@handle_abort
@init_cache
@attach_settings
def update(
    ctx: typer.Context,
    openai_api_key: Annotated[
        str | None, typer.Option(help="The API key needed to access OpenAI")
    ] = None,
    resume_path: Annotated[
        Path | None, typer.Option(help="The path to your resume")
    ] = None,
    candidate_name: Annotated[
        str | None, typer.Option(help="The name of the candidate to use in the closing")
    ] = None,
    filename_prefix: Annotated[
        str | None,
        typer.Option(help="The filename prefix to use for your cover letter"),
    ] = None,
    sig_path: Annotated[
        Path | None, typer.Option(help="The path to your signature")
    ] = None,
    heading_path: Annotated[
        Path | None, typer.Option(help="The path to your markdown heading")
    ] = None,
):
    """
    Bind the configuration to the app.
    """
    logger.debug(f"Updating settings with {locals()}")
    kwargs: dict[str, Any] = {}
    if openai_api_key is not None:
        kwargs["openai_api_key"] = openai_api_key
    if resume_path is not None:
        kwargs["resume_path"] = resume_path
    if candidate_name is not None:
        kwargs["candidate_name"] = candidate_name
    if filename_prefix is not None:
        kwargs["filename_prefix"] = filename_prefix
    if sig_path is not None:
        kwargs["sig_path"] = sig_path
    if heading_path is not None:
        kwargs["heading_path"] = heading_path

    settings = update_settings(ctx.obj.settings, **kwargs)
    dump_settings(settings)


@cli.command()
@handle_abort
@init_cache
@attach_settings
def unset(
    ctx: typer.Context,
    openai_api_key: Annotated[
        bool, typer.Option(help="The API key needed to access OpenAI")
    ] = False,
    resume_path: Annotated[bool, typer.Option(help="The path to your resume")] = False,
    candidate_name: Annotated[
        bool, typer.Option(help="The name of the candidate to use in the closing")
    ] = False,
    filename_prefix: Annotated[
        bool, typer.Option(help="The filename prefix to use for your cover letter")
    ] = False,
    sig_path: Annotated[bool, typer.Option(help="The path to your signature")] = False,
    heading_path: Annotated[
        bool, typer.Option(help="The path to your markdown heading")
    ] = False,
):
    """
    Bind the configuration to the app.
    """
    logger.debug(f"Updating settings with {locals()}")
    keys = [k for k in locals() if locals()[k]]
    settings = unset_settings(ctx.obj.settings, *keys)
    dump_settings(settings)


@cli.command()
@handle_abort
@init_cache
@attach_settings
def show(ctx: typer.Context):
    """
    Show the config that is currently bound to the app.
    """
    parts = []
    for field_name, field_value in ctx.obj.settings:
        parts.append((field_name, field_value))
    max_field_len = max(len(field_name) for field_name, _ in parts)
    message = "\n".join(f"[bold]{k:<{max_field_len}}[/bold] -> {v}" for k, v in parts)
    terminal_message(message, subject="Current Configuration")


@cli.command()
@handle_abort
@init_cache
def clear():
    """
    Clear the config from the app.
    """
    logger.debug("Clearing settings")
    clear_settings()



from functools import wraps
from pathlib import Path

from smart_letters.exceptions import Abort


CACHE_DIR: Path = Path.home() / ".local/share/smart-letters"


def init_cache(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            CACHE_DIR.mkdir(exist_ok=True, parents=True)
            info_file = CACHE_DIR / "info.txt"
            info_file.write_text(
                "This directory is used by Smart Letters for its cache."
            )
        except Exception:
            raise Abort(
                """
                Cache directory {cache_dir} doesn't exist, is not writable,
                or could not be created.

                Please check your home directory permissions and try again.
                """,
                subject="Non-writable cache dir",
                log_message="Non-writable cache dir",
            )
        return func(*args, **kwargs)

    return wrapper



import snick
import typer

from smart_letters.config import cli as config_cli
from smart_letters.format import terminal_message
from smart_letters.generate import cli as generate_cli
from smart_letters.logging import init_logs
from smart_letters.schemas import CliContext
from smart_letters.version import show_version


cli = typer.Typer(rich_markup_mode="rich")


@cli.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    verbose: bool = typer.Option(False, help="Enable verbose logging to the terminal"),
    version: bool = typer.Option(False, help="Print the version of this app and exit"),
):
    """
    Welcome to Smart Letters!

    More information can be shown for each command listed below by running it with the
    --help option.
    """

    if version:
        show_version()
        ctx.exit()

    if ctx.invoked_subcommand is None:
        terminal_message(
            snick.conjoin(
                "No command provided. Please check [bold magenta]usage[/bold magenta]",
                "",
                f"[yellow]{ctx.get_help()}[/yellow]",
            ),
            subject="Need an Armasec command",
        )
        ctx.exit()

    init_logs(verbose=verbose)
    ctx.obj = CliContext()


cli.add_typer(config_cli, name="config")
cli.add_typer(generate_cli, name="generate")


if __name__ == "__main__":
    cli()



from typing import Any

import snick
from rich.console import Console
from rich.panel import Panel
from rich.markdown import Markdown


def terminal_message(
    message,
    subject=None,
    color="green",
    footer=None,
    indent=True,
    markdown=False,
):
    panel_kwargs: dict[str, Any] = dict(padding=1)
    if subject is not None:
        panel_kwargs["title"] = f"[{color}]{subject}"
    if footer is not None:
        panel_kwargs["subtitle"] = f"[dim italic]{footer}[/dim italic]"
    text = snick.dedent(message)
    if indent:
        text = snick.indent(text, prefix="  ")
    if markdown:
        text = Markdown(text)
    console = Console()
    console.print()
    console.print(Panel(text, **panel_kwargs))
    console.print()


def simple_message(message, indent=False, markdown=False):
    text = snick.dedent(message)
    if indent:
        text = snick.indent(text, prefix="  ")
    if markdown:
        text = Markdown(text)
    console = Console()
    console.print()
    console.print(text)
    console.print()
